==> wyvern-2.2.7.tar.bz2: wyvern-2.2.7/modules/mod_access.c <==
/*****************************************************************************
 *
 * FILE:	mod_access.c
 * DESCRIPTION:	Wyvern: access control module
 * DATE:	Sat, Sep 16 2000
 * UPDATED:	Wed, Mar  1 2006
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2000-2006 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2000-2006 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_access.c,v 1.28 2006/03/01 13:21:37 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#include "wyvern.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	WM_MODULE_ID	"mod_access"

#define	HOST_NAME_ANY	"*"

typedef enum {
  ACCESS_DENY	= 0,
  ACCESS_ALLOW	= 1
} wm_access_t;

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	wm_access_init(void);
static void	wm_access_finish(void);
static int	wm_access_checker(Message *);
static int	access_checker(Message *, sxml_node_t *, wm_access_t);
static int	compare_domain(const char *, const char *);
static int	compare_address(const char *, const char *);
static void	wm_access_register_hooks(void);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static sxml_node_t *	root = NULL;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wm_access_init
 * Description:		initialize the host access list
 * Arguments:		none
 * Return Value:	none
 */
static void
wm_access_init(void)
{
  char *		path;
  register FILE *	fin;

  /* read 'ServerRoot/conf/waccess.conf' */
  ws_asprintf(&path, "%s/%s", httpd->configDir, WYVERN_SERVER_ACCESS);

  /* register allowed access hosts */
  fin = ws_safe_fopen(path, "r");	free(path);
  if (fin != NULL) {
    root = sxml_parse_file(fileno(fin));
    if (root != NULL) {
      sxml_node_t *	node;

      node = sxml_find_element(root, "wyvern-access", NULL, NULL);
      if (node == NULL) {	/* error: invalid formatted waccess.conf */
	wm_access_finish();
      }
    }
    fclose(fin);
  }
}

/*
 * Function Name:	wm_access_finish
 * Description:		release the memory area of access lists
 * Arguments:		none
 * Return Value:	none
 */
void
wm_access_finish(void)
{
  if (root != NULL) {
    sxml_delete_node(root);
    root = NULL;
  }
}

/*
 * Function Name:	wm_access_checker
 * Description:		permit the access to the host
 * Arguments:		message
 * Return Value:	allowed host: 0, denied host: -1
 */
static int
wm_access_checker(msg)
	register Message *	msg;
{
  if (root != NULL) {
    register sxml_node_t *	np;

    for (np = sxml_find_element(root, "host", "name", NULL);
	 np != NULL; np = np->next) {
      const char *	hostname;
      int		retval = WM_RET_OK;

      hostname = sxml_get_attribute(np, "name");
#if	DEBUG
      fprintf(stderr, "[%s] %s<=>%s\n", WM_MODULE_ID, hostname, msg->vhost->serverName);
#endif	/* DEBUG */
      if (ws_strcmp(hostname, msg->vhost->serverName) == 0 ||
	  ws_strcmp(hostname, HOST_NAME_ANY) == 0) {
	const char *		defval;
	register sxml_node_t *	nnp;

	defval = sxml_get_attribute(np, "default");
	retval = ws_strcmp(defval, "deny") == 0 ? WM_RET_ERROR : WM_RET_OK;
	for (nnp = sxml_find_element(np->child, "limit", "directory", NULL);
	     nnp != NULL; nnp = nnp->next) {
	  const char *	dirname;

	  dirname = sxml_get_attribute(nnp, "directory");
	  if (dirname != NULL && strstr(msg->request.uri, dirname) != NULL) {
	    return access_checker(msg, nnp,
				  retval == WM_RET_ERROR ? ACCESS_DENY
							 : ACCESS_ALLOW);
	  }
	}
      }
      return retval;
    }
  }

  return WM_RET_OK;
}

static int
access_checker(msg, tree, defval)
	register Message *	msg;
	register sxml_node_t *	tree;
	wm_access_t		defval;	/* allow | deny */
{
  struct sockaddr	sa;
  char			hbuf[NI_MAXHOST];
  sxml_node_t *		node;

  DBG(fputs("<access_checker>\n", stderr);)

  /* get client hostname with DNS */
  memset(hbuf, 0, sizeof(hbuf));
  memcpy(&sa, msg->client.addr, sizeof(struct sockaddr));
#if	HAVE_SOCKADDR_SA_LEN
  if (getnameinfo(&sa, sa.sa_len, hbuf, sizeof(hbuf), NULL, 0,
		  NI_NAMEREQD) != 0) {
    wv_syslog(WLL_LOG_SECURITY, "[%s] No canonical name: %s",
	      WM_MODULE_ID, msg->client.ip);
    return WM_RET_ERROR;
  }
#else
  if (getnameinfo(&sa, sizeof(sa), hbuf, sizeof(hbuf), NULL, 0,
		  NI_NAMEREQD) != 0) {
    wv_syslog(WLL_LOG_SECURITY, "[%s] No canonical name: %s",
	      WM_MODULE_ID, msg->client.ip);
    return WM_RET_ERROR;
  }
#endif	/* HAVE_SOCKADDR_SA_LEN */
  msg->client.name = ws_strdup(hbuf);

  if (defval == ACCESS_ALLOW) {
    node = sxml_find_element(tree->child, "deny", NULL, NULL);
  }
  else {
    node = sxml_find_element(tree->child, "allow", NULL, NULL);
  }
  if (node != NULL) {
    register sxml_node_t *	np;

    for (np = node->child; np != NULL; np = np->next) {
      const char *	content;
      int		retval = -1;

      if (np->type != SXML_ELEMENT)	  { continue; }
      if (np->value.element.name == NULL) { continue; }

      /*
       * NOTES: This access module will apply the first matched <address> or
       * <domain> tag. 
       */
      content = sxml_get_content(np->child);
      if (ws_strcmp(np->value.element.name, "address") == 0) {
	retval = compare_address(msg->client.ip, content);
      }
      else if (ws_strcmp(np->value.element.name, "domain") == 0) {
	retval = compare_domain(msg->client.name, content);
      }
      else {
	wv_syslog(WLL_LOG_NOTICE, "[%s] unknown tag: %s",
		  WM_MODULE_ID, np->value.element.name);
      }
      if (retval == 0) {	/* matched */
	/* return reverse value of default access */
	return defval ? WM_RET_ERROR : WM_RET_OK;
      }
    }
  }

  return WM_RET_OK;
}

static int
compare_domain(name, pat)
	register const char *	name;
	register const char *	pat;
{
  size_t	nlen;
  size_t	plen;

  if (name == NULL || pat == NULL) {
    return -1;
  }

  DBG(fprintf(stderr, "[compare_domain] %s<=>%s\n", name, pat);)
  nlen = strlen(name);
  plen = strlen(pat);
  if (nlen < plen) {
    return -1;
  }

  /* check to match the name with the pattern like as '.ad.jp' */
  if (nlen >= plen && pat[0] == '.' &&
      ws_strncmp(name + nlen - plen, pat, plen) == 0) {
    return 0;	/* match! */
  }
  /* check to match the name exactly */
  if (ws_strcmp(name, pat) == 0) {
    return 0;	/* match! */
  }

  return -1;	/* no match */
}

/*****************************************************************************/

typedef union {
  in_addr_t	bits;
  uint8_t	a[4];
} wv_netmask_t;

typedef struct {
  uint8_t	inet[4];
  wv_netmask_t	mask;
} wv_ipaddr_t;

static int	parse_addr(const char *, wv_ipaddr_t *);
static int	parse_cidr(const char *, wv_ipaddr_t **);
static int	check_addr(const char *, wv_ipaddr_t *);

static int
parse_addr(ipv4, retval)
	const char *	ipv4;
	wv_ipaddr_t *	retval;
{
  char *	s;
  char **	ap;
  char *	args[4];
  int		n;
  register int	i;
  wv_netmask_t	x;

  s = ws_strdup(ipv4);
  for (n = 0, ap = args; (*ap = strsep(&s, ".")) != NULL; ) {
    if (**ap != '\0') {
      n++;
      if (++ap >= &args[4]) {
	break;
      }
    }
  }

  x.bits = htonl(retval->mask.bits);
  for (i = 0; i < n; i++) {
    retval->inet[i] = (uint8_t)strtol(args[i], (char **)NULL, 10) & x.a[i];
  }
  free(s);

  return 0;
}

static int
parse_cidr(ipv4, retval)
	const char *	ipv4;
	wv_ipaddr_t **	retval;
{
  char *		ap;
  register char *	cp;
  int			cidr = 32;
  in_addr_t		mask = 0xffffffff;

  ap = ws_strdup(ipv4);
  cp = strchr(ap, '/');
  if (cp != NULL) {
    *cp++ = '\0';
    cidr = (int)strtol(cp, (char **)NULL, 10);
  }
  mask <<= (32 - cidr);

  if (retval != NULL) {
    *retval		 = (wv_ipaddr_t *)ws_malloc(sizeof(wv_ipaddr_t));
    (*retval)->mask.bits = mask;
    if (parse_addr(ap, *retval) == -1) {
      free(*retval);
      *retval = NULL;
    }
  }
  free(ap);

  return (*retval != NULL) ? 0 : -1;
}

static int
check_addr(addr, net)
	const char *	addr;
	wv_ipaddr_t *	net;
{
  wv_ipaddr_t *	ip;
  int		retval = 0;

  if (parse_cidr(addr, &ip) == 0) {
    register int	i;
    wv_netmask_t	x;

    x.bits = htonl(net->mask.bits);
    for (i = 0; i < 4; i++) {
      if (net->inet[i] != (ip->inet[i] & x.a[i])) {
	retval = -1;
      }
    }
    free(ip);
  }

  return retval;
}

static int
compare_address(name, pat)
	register const char *	name;
	register const char *	pat;
{
  wv_ipaddr_t *	addr;
  int		retval = -1;

  DBG(fprintf(stderr, "[compare_address] %s<=>%s\n", name, pat);)
  if (parse_cidr(pat, &addr) == 0) {
    retval = check_addr(name, addr);
#if	DEBUG
    fprintf(stderr, "[compare_address] %s<=>%d.%d.%d.%d/%x [%d]\n",
	    name,
	    addr->inet[0], addr->inet[1], addr->inet[2], addr->inet[3],
	    addr->mask.bits, retval);
#endif	/* DEBUG */
    free(addr);
  }

  return retval;
}

/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_access_register_hooks(void)
{
  wm_hook_access_checker(wm_access_checker, HOOK_MIDDLE);
}

static wmModule	mod_access = {
  WM_MAGIC_NUMBER_MAJOR,			/* version */
  WM_SECURE,					/* secure */
  WM_MODULE_ID,					/* name */
  "Access Control Module version 1.2.5",	/* info */
  "Kouichi ABE (WALL)",				/* author */
  NULL,						/* depend */
  NULL,						/* handle */
  (void *)wm_access_init,			/* init */
  (void *)wm_access_finish,			/* finish */
  (void *)wm_access_register_hooks		/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &mod_access;
}

==> wyvern-2.2.10.tar.bz2: wyvern-2.2.10/modules/mod_access.c <==
/*****************************************************************************
 *
 * FILE:	mod_access.c
 * DESCRIPTION:	Wyvern: access control module
 * DATE:	Sat, Sep 16 2000
 * UPDATED:	Sat, Aug 25 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2000-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2000-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_access.c,v 1.29 2007/08/24 18:01:55 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#include "wyvern.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	WM_MODULE_ID	"mod_access"

#define	HOST_NAME_ANY	"*"

typedef enum {
  ACCESS_DENY	= 0,
  ACCESS_ALLOW	= 1
} wm_access_t;

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	wm_access_init(void);
static void	wm_access_finish(void);
static int	wm_access_checker(Message *);
static int	access_checker(Message *, sxml_node_t *, wm_access_t);
static int	compare_domain(const char *, const char *);
static int	compare_address(const char *, const char *);
static void	wm_access_register_hooks(void);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static sxml_node_t *	root = NULL;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wm_access_init
 * Description:		initialize the host access list
 * Arguments:		none
 * Return Value:	none
 */
static void
wm_access_init(void)
{
  char *		path;
  register FILE *	fin;

  /* read 'ServerRoot/conf/waccess.conf' */
  ws_asprintf(&path, "%s/%s", httpd->configDir, WYVERN_SERVER_ACCESS);

  /* register allowed access hosts */
  fin = ws_safe_fopen(path, "r");	free(path);
  if (fin != NULL) {
    root = sxml_parse_file(fileno(fin));
    if (root != NULL) {
      sxml_node_t *	node;

      node = sxml_find_element(root, "wyvern-access", NULL, NULL);
      if (node == NULL) {	/* error: invalid formatted waccess.conf */
	wm_access_finish();
      }
    }
    fclose(fin);
  }
}

/*
 * Function Name:	wm_access_finish
 * Description:		release the memory area of access lists
 * Arguments:		none
 * Return Value:	none
 */
void
wm_access_finish(void)
{
  if (root != NULL) {
    sxml_delete_node(root);
    root = NULL;
  }
}

/*
 * Function Name:	wm_access_checker
 * Description:		permit the access to the host
 * Arguments:		message
 * Return Value:	allowed host: 0, denied host: -1
 */
static int
wm_access_checker(msg)
	register Message *	msg;
{
  if (root != NULL) {
    register sxml_node_t *	np;

    for (np = sxml_find_element(root, "host", "name", NULL);
	 np != NULL; np = np->next) {
      const char *	hostname;
      int		status = WM_RET_OK;

      hostname = sxml_get_attribute(np, "name");
#if	DEBUG
      fprintf(stderr, "[%s] %s<=>%s\n", WM_MODULE_ID, hostname, msg->vhost->serverName);
#endif	/* DEBUG */
      if (ws_strcmp(hostname, msg->vhost->serverName) == 0 ||
	  ws_strcmp(hostname, HOST_NAME_ANY) == 0) {
	const char *		defval;
	register sxml_node_t *	nnp;

	defval = sxml_get_attribute(np, "default");
	status = ws_strcmp(defval, "deny") == 0 ? WM_RET_ERROR : WM_RET_OK;
	for (nnp = sxml_find_element(np->child, "limit", "directory", NULL);
	     nnp != NULL; nnp = nnp->next) {
	  const char *	dirname;

	  dirname = sxml_get_attribute(nnp, "directory");
	  if (dirname != NULL && strstr(msg->request.uri, dirname) != NULL) {
	    return access_checker(msg, nnp,
				  status == WM_RET_ERROR ? ACCESS_DENY
							 : ACCESS_ALLOW);
	  }
	}
      }
      return status;
    }
  }

  return WM_RET_OK;
}

static int
access_checker(msg, tree, defval)
	register Message *	msg;
	register sxml_node_t *	tree;
	wm_access_t		defval;	/* allow | deny */
{
  struct sockaddr	sa;
  char			hbuf[NI_MAXHOST];
  sxml_node_t *		node;

  DBG(fputs("<access_checker>\n", stderr);)

  /* get client hostname with DNS */
  memset(hbuf, 0, sizeof(hbuf));
  memcpy(&sa, msg->client.addr, sizeof(struct sockaddr));
#if	HAVE_SOCKADDR_SA_LEN
  if (getnameinfo(&sa, sa.sa_len, hbuf, sizeof(hbuf), NULL, 0,
		  NI_NAMEREQD) != 0) {
    wv_syslog(WLL_LOG_SECURITY, "[%s] No canonical name: %s",
	      WM_MODULE_ID, msg->client.ip);
    return WM_RET_ERROR;
  }
#else
  if (getnameinfo(&sa, sizeof(sa), hbuf, sizeof(hbuf), NULL, 0,
		  NI_NAMEREQD) != 0) {
    wv_syslog(WLL_LOG_SECURITY, "[%s] No canonical name: %s",
	      WM_MODULE_ID, msg->client.ip);
    return WM_RET_ERROR;
  }
#endif	/* HAVE_SOCKADDR_SA_LEN */
  msg->client.name = ws_strdup(hbuf);

  if (defval == ACCESS_ALLOW) {
    node = sxml_find_element(tree->child, "deny", NULL, NULL);
  }
  else {
    node = sxml_find_element(tree->child, "allow", NULL, NULL);
  }
  if (node != NULL) {
    register sxml_node_t *	np;

    for (np = node->child; np != NULL; np = np->next) {
      const char *	content;
      int		status = -1;

      if (np->type != SXML_ELEMENT)	  { continue; }
      if (np->value.element.name == NULL) { continue; }

      /*
       * NOTES: This access module will apply the first matched <address> or
       * <domain> tag. 
       */
      content = sxml_get_content(np->child);
      if (ws_strcmp(np->value.element.name, "address") == 0) {
	status = compare_address(msg->client.ip, content);
      }
      else if (ws_strcmp(np->value.element.name, "domain") == 0) {
	status = compare_domain(msg->client.name, content);
      }
      else {
	wv_syslog(WLL_LOG_NOTICE, "[%s] unknown tag: %s",
		  WM_MODULE_ID, np->value.element.name);
      }
      if (status == 0) {	/* matched */
	/* return reverse value of default access */
	return defval ? WM_RET_ERROR : WM_RET_OK;
      }
    }
  }

  return WM_RET_OK;
}

static int
compare_domain(name, pat)
	register const char *	name;
	register const char *	pat;
{
  size_t	nlen;
  size_t	plen;

  if (name == NULL || pat == NULL) {
    return -1;
  }

  DBG(fprintf(stderr, "[compare_domain] %s<=>%s\n", name, pat);)
  nlen = strlen(name);
  plen = strlen(pat);
  if (nlen < plen) {
    return -1;
  }

  /* check to match the name with the pattern like as '.ad.jp' */
  if (nlen >= plen && pat[0] == '.' &&
      ws_strncmp(name + nlen - plen, pat, plen) == 0) {
    return 0;	/* match! */
  }
  /* check to match the name exactly */
  if (ws_strcmp(name, pat) == 0) {
    return 0;	/* match! */
  }

  return -1;	/* no match */
}

/*****************************************************************************/

typedef union {
  in_addr_t	bits;
  uint8_t	a[4];
} wv_netmask_t;

typedef struct {
  uint8_t	inet[4];
  wv_netmask_t	mask;
} wv_ipaddr_t;

static int	parse_addr(const char *, wv_ipaddr_t *);
static int	parse_cidr(const char *, wv_ipaddr_t **);
static int	check_addr(const char *, wv_ipaddr_t *);

static int
parse_addr(ipv4, retval)
	const char *	ipv4;
	wv_ipaddr_t *	retval;
{
  char *	s;
  char **	ap;
  char *	args[4];
  int		n;
  register int	i;
  wv_netmask_t	x;

  s = ws_strdup(ipv4);
  for (n = 0, ap = args; (*ap = strsep(&s, ".")) != NULL; ) {
    if (**ap != '\0') {
      n++;
      if (++ap >= &args[4]) {
	break;
      }
    }
  }

  x.bits = htonl(retval->mask.bits);
  for (i = 0; i < n; i++) {
    retval->inet[i] = (uint8_t)strtol(args[i], (char **)NULL, 10) & x.a[i];
  }
  free(s);

  return 0;
}

static int
parse_cidr(ipv4, retval)
	const char *	ipv4;
	wv_ipaddr_t **	retval;
{
  char *		ap;
  register char *	cp;
  int			cidr = 32;
  in_addr_t		mask = 0xffffffff;

  ap = ws_strdup(ipv4);
  cp = strchr(ap, '/');
  if (cp != NULL) {
    *cp++ = '\0';
    cidr = (int)strtol(cp, (char **)NULL, 10);
  }
  mask <<= (32 - cidr);

  if (retval != NULL) {
    *retval		 = (wv_ipaddr_t *)ws_malloc(sizeof(wv_ipaddr_t));
    (*retval)->mask.bits = mask;
    if (parse_addr(ap, *retval) == -1) {
      free(*retval);
      *retval = NULL;
    }
  }
  free(ap);

  return (*retval != NULL) ? 0 : -1;
}

static int
check_addr(addr, net)
	const char *	addr;
	wv_ipaddr_t *	net;
{
  wv_ipaddr_t *	ip;
  int		status = 0;

  if (parse_cidr(addr, &ip) == 0) {
    register int	i;
    wv_netmask_t	x;

    x.bits = htonl(net->mask.bits);
    for (i = 0; i < 4; i++) {
      if (net->inet[i] != (ip->inet[i] & x.a[i])) {
	status = -1;
      }
    }
    free(ip);
  }

  return status;
}

static int
compare_address(name, pat)
	register const char *	name;
	register const char *	pat;
{
  wv_ipaddr_t *	addr;
  int		status = -1;

  DBG(fprintf(stderr, "[compare_address] %s<=>%s\n", name, pat);)
  if (parse_cidr(pat, &addr) == 0) {
    status = check_addr(name, addr);
#if	DEBUG
    fprintf(stderr, "[compare_address] %s<=>%d.%d.%d.%d/%x [%d]\n",
	    name,
	    addr->inet[0], addr->inet[1], addr->inet[2], addr->inet[3],
	    addr->mask.bits, status);
#endif	/* DEBUG */
    free(addr);
  }

  return status;
}

/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_access_register_hooks(void)
{
  wm_hook_access_checker(wm_access_checker, HOOK_MIDDLE);
}

static wmModule	mod_access = {
  WM_MAGIC_NUMBER_MAJOR,			/* version */
  WM_SECURE,					/* secure */
  WM_MODULE_ID,					/* name */
  "Access Control Module version 1.2.5",	/* info */
  "Kouichi ABE (WALL)",				/* author */
  NULL,						/* depend */
  NULL,						/* handle */
  (void *)wm_access_init,			/* init */
  (void *)wm_access_finish,			/* finish */
  (void *)wm_access_register_hooks		/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &mod_access;
}
